/**
 * Controller class for interacting with a HAP device over GATT.
 */
/// <reference types="node" />
import { EventEmitter } from 'events';
import GattConnection from './gatt-connection';
import { PairingData } from '../../protocol/pairing-protocol';
import { TLV } from '../../model/tlv';
import { Characteristic as NobleCharacteristic, Peripheral as NoblePeripheral } from '@abandonware/noble';
export interface GetCharacteristicsOptions {
    meta?: boolean;
    perms?: boolean;
    type?: boolean;
    ev?: boolean;
    extra?: boolean;
}
export interface CharacteristicObject {
    serviceUuid?: string;
    aid: number;
    iid: number;
    type?: string;
    value?: unknown;
    perms?: string[];
    ev?: boolean;
    description?: string;
    format?: string;
    unit?: string;
    minValue?: number;
    maxValue?: number;
    minStep?: number;
    maxLen?: number;
    maxDataLen?: number;
    'valid-values'?: number[];
    'valid-values-range'?: number[];
    TTL?: number;
    pid?: number;
}
export interface Accessories {
    accessories: {
        aid: number;
        services: {
            iid: number;
            type: string;
            characteristics: {
                type: string;
                ev: boolean;
                perms: string[];
                format: string;
                iid?: number;
            }[];
            primary?: boolean;
            hidden?: boolean;
        }[];
    }[];
}
export default class GattClient extends EventEmitter {
    deviceId: string;
    private peripheral;
    private pairingProtocol;
    private gattProtocol;
    private tid;
    private queue;
    private _pairingConnection?;
    /**
     * Initialize the GattClient object.
     *
     * @param {string} deviceId - ID of the device
     * @param {Object} peripheral - Peripheral object from noble
     * @param {Object?} pairingData - existing pairing data
     */
    constructor(deviceId: string, peripheral: NoblePeripheral, pairingData?: PairingData);
    /**
     * Queue an operation for the client.
     *
     * @param {function} op - Function to add to the queue
     * @returns {Promise} Promise which resolves when the function is called.
     */
    private _queueOperation;
    /**
     * Get the next transaction ID.
     *
     * @returns {number} Transaction ID.
     */
    getNextTransactionId(): number;
    /**
     * Get the data (keys) that needs to be stored long-term.
     *
     * @returns {Object} Object containing the keys that should be stored.
     */
    getLongTermData(): PairingData | null;
    /**
     * Run the identify routine on a device.
     *
     * @returns {Promise} Promise which resolves if identify succeeded.
     */
    identify(): Promise<void>;
    /**
     * Read the instance ID descriptor for a characteristic. The peripheral must
     * already be connected.
     *
     * @param {Object} characteristic - The characteristic to read from
     * @returns {Promise} Promise which resolves to the IID.
     */
    private _readInstanceId;
    /**
     * Begins the pairing process. For devices with random pins, this
     * will cause it to show the pin on the screen.
     *
     * @returns {Promise} Promise which resolves to opaque
     * pairing data when complete.
     */
    startPairing(): Promise<{
        tlv: TLV;
        iid: number;
        characteristic: NobleCharacteristic;
    }>;
    /**
     * Finishes a pairing process that began with startPairing()
     *
     * @param {Object} pairingData - The pairing data returned from startPairing()
     * @returns {Promise} Promise which resolves when pairing is complete.
     */
    finishPairing(pairingData: {
        tlv: TLV;
        iid: number;
        characteristic: NobleCharacteristic;
    }, pin: string): Promise<void>;
    /**
     * Attempt to pair with a device.
     *
     * @param {string} pin - The pairing PIN
     * @returns {Promise} Promise which resolves when pairing is complete.
     */
    pairSetup(pin: string): Promise<void>;
    /**
     * Method used internally to generate session keys for a connection.
     *
     * @param {Object} connection - Existing GattConnection object
     * @returns {Promise} Promise which resolves when the pairing has been verified.
     */
    private _pairVerify;
    /**
     * Unpair the controller from a device.
     *
     * @param {string} identifier - Identifier of the controller to remove
     * @returns {Promise} Promise which resolves when the process completes.
     */
    removePairing(identifier: string): Promise<void>;
    /**
     * Add a pairing to a device.
     *
     * @param {string} identifier - Identifier of new controller
     * @param {Buffer} ltpk - Long-term public key of the new controller
     * @param {boolean} isAdmin - Whether or not the new controller is an admin
     * @returns {Promise} Promise which resolves when the process is complete.
     */
    addPairing(identifier: string, ltpk: Buffer, isAdmin: boolean): Promise<void>;
    /**
     * List the pairings on a device.
     *
     * @returns {Promise} Promise which resolves to the final TLV when the process
     *                    is complete.
     */
    listPairings(): Promise<TLV>;
    /**
     * Get the accessory attribute database from a device.
     *
     * @returns {Promise} Promise which resolves to the JSON document.
     */
    getAccessories(): Promise<Accessories>;
    /**
     * Read a set of characteristics.
     *
     * @param {Object[]} characteristics - Characteristics to get, as a list of
     *                   objects: {characteristicUuid, serviceUuid, iid, format}
     * @param {Object?} options - Options dictating what metadata to fetch
     * @param {Object?} connection - Existing GattConnection object, must already
     *                  be paired and verified
     * @returns {Promise} Promise which resolves to the JSON document.
     */
    getCharacteristics(characteristics: {
        characteristicUuid: string;
        serviceUuid: string;
        iid: number;
        format?: string;
    }[], options?: GetCharacteristicsOptions, connection?: GattConnection | null): Promise<{
        characteristics: CharacteristicObject[];
    }>;
    /**
     * Modify a set of characteristics.
     *
     * @param {Object[]} values - Characteristics to set, as a list of objects:
     *                   {characteristicUuid, serviceUuid, iid, value}
     * @returns {Promise} Promise which resolves when the characteristics have been set.
     */
    setCharacteristics(values: {
        characteristicUuid: string;
        serviceUuid: string;
        iid: number;
        value: unknown;
    }[]): Promise<void>;
    /**
     * Subscribe to events for a set of characteristics.
     *
     * @param {Object[]} characteristics - Characteristics to subscribe to, as a
     *                   list of objects:
     *                   {characteristicUuid, serviceUuid, iid, format}
     * @returns {Promise} Promise which resolves to the GattConnection object.
     */
    subscribeCharacteristics(characteristics: {
        characteristicUuid: string;
        serviceUuid: string;
        iid: number;
        format: string;
    }[]): Promise<GattConnection>;
    /**
     * Unsubscribe from events for a set of characteristics.
     *
     * @param {Object[]} characteristics - Characteristics to unsubscribe from, as
     *                   a list of objects: {characteristicUuid, serviceUuid}
     * @returns {Promise} Promise which resolves when the procedure is done.
     */
    unsubscribeCharacteristics(characteristics: {
        characteristicUuid: string;
        serviceUuid: string;
    }[]): Promise<void>;
}
//# sourceMappingURL=gatt-client.d.ts.map